﻿Imports System.ComponentModel
Imports System.Text

Public Class MultiColumnComboBox

    Private WithEvents Value_Titles As New TitleItemCollection
    Private WithEvents Value_Values As New ValueItemCollection
    Private Value_DropDownWidth As Integer = 300
    Private Value_DropDownHeight As Integer = 200

    Public Event SelectedItemChanged(ByVal sender As Object, ByVal e As System.EventArgs)
    Public Event SelectedItemChangedByUser(ByVal sender As Object, ByVal e As System.EventArgs)
    Public Event DropdownRequested(ByVal sender As Object, ByVal e As DropdownRequestEventArgs)

#Region "Properties"

    Public ReadOnly Property Titles() As TitleItemCollection
        Get
            Return Value_Titles
        End Get
    End Property

    Public ReadOnly Property Values() As ValueItemCollection
        Get
            Return Value_Values
        End Get
    End Property

    <Description("Width of dropdown window"), Category("Dropdown")> _
    Public Property DropDownWindow_Width() As Integer
        Get
            Return Value_DropDownWidth
        End Get
        Set(ByVal value As Integer)
            Value_DropDownWidth = value
        End Set
    End Property

    <Description("Height of dropdown window"), Category("Dropdown")> _
    Public Property DropDownWindow_Height() As Integer
        Get
            Return Value_DropDownHeight
        End Get
        Set(ByVal value As Integer)
            Value_DropDownHeight = value
        End Set
    End Property

    Public Overrides Property BackColor() As System.Drawing.Color
        Get
            Return TextBox_Display.BackColor
        End Get
        Set(ByVal value As System.Drawing.Color)
            TextBox_Display.BackColor = value
        End Set
    End Property

    Public Shadows Property BorderStyle() As System.Windows.Forms.BorderStyle
        Get
            Return TextBox_Display.BorderStyle
        End Get
        Set(ByVal value As System.Windows.Forms.BorderStyle)
            TextBox_Display.BorderStyle = value
        End Set
    End Property

    Public Overrides Property Cursor() As System.Windows.Forms.Cursor
        Get
            Return TextBox_Display.Cursor
        End Get
        Set(ByVal value As System.Windows.Forms.Cursor)
            TextBox_Display.Cursor = value
        End Set
    End Property

    Public Overrides Property ForeColor() As System.Drawing.Color
        Get
            Return TextBox_Display.ForeColor
        End Get
        Set(ByVal value As System.Drawing.Color)
            TextBox_Display.ForeColor = value
        End Set
    End Property

    Public Shadows ReadOnly Property Text() As String
        Get
            Return TextBox_Display.Text
        End Get
    End Property

    <Description("Indicates how the text should be aligned for this control."), Category("Appearance")> _
    Public Property TextAlign() As System.Windows.Forms.HorizontalAlignment
        Get
            Return TextBox_Display.TextAlign
        End Get
        Set(ByVal value As System.Windows.Forms.HorizontalAlignment)
            TextBox_Display.TextAlign = value
        End Set
    End Property

    Public Shadows Property UseWaitCursor() As Boolean
        Get
            Return TextBox_Display.UseWaitCursor
        End Get
        Set(ByVal value As Boolean)
            TextBox_Display.UseWaitCursor = value
        End Set
    End Property

    Private Sub MultiColumnComboBox_Disposed(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Disposed
        If SubWindow IsNot Nothing Then CloseSubWindow()
    End Sub

    Private Sub MultiColumnComboBox_FontChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.FontChanged
        Height = TextBox_Display.Height
        Width = TextBox_Display.Width
    End Sub

    Private Sub MultiColumnComboBox_Resize(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Resize
        Height = TextBox_Display.Height
    End Sub

#End Region

#Region "SelectedID and Text"

    Private Value_TextMask As String = "[0]"

    <Description("Define the mask of text."), Category("Appearance")> _
    Public Property TextMask() As String
        Get
            Return Value_TextMask
        End Get
        Set(ByVal value As String)
            Value_TextMask = value
        End Set
    End Property

    Private Value_SelectedID As Object

    <Description("Get or set the ID of selected item."), Category("Data")> _
    Public Property SelectedID() As Object
        Get
            Return Value_SelectedID
        End Get
        Set(ByVal value As Object)
            If value <> Value_SelectedID Then
                Value_SelectedID = value
                ApplySelectText(value)
                CloseSubWindow()
                RaiseEvent SelectedItemChanged(Me, New EventArgs)
            Else
                ApplySelectText(value)
            End If
        End Set
    End Property

    Private Sub ApplySelectText(ByVal ID As Object)
        If ID Is Nothing OrElse Value_TextMask = "" Then
            TextBox_Display.Text = ""
            Return
        End If

        Dim Final As New StringBuilder
        Dim Para As StringBuilder = Nothing
        Dim Index As Integer
        Dim ParaMode As Boolean = False
        Dim MaskArray As Char() = Value_TextMask.ToCharArray
        For Index = 0 To Value_TextMask.Length - 1
            If MaskArray(Index) = "[" Then
                ' [[
                If Index < Value_TextMask.Length - 1 Then
                    If MaskArray(Index + 1) = "[" Then
                        If ParaMode Then
                            Para.Append("[")
                        Else
                            Final.Append("[")
                        End If
                        Index += 1
                        Continue For
                    End If
                End If

                If ParaMode Then
                    Final.Append("[")
                    Final.Append(Para.ToString)
                Else
                    ParaMode = True
                End If
                Para = New StringBuilder
            ElseIf MaskArray(Index) = "]" Then
                If ParaMode Then
                    Dim ParaString As String = Para.ToString()
                    Dim ParaIndex As Integer
                    Dim Line As String() = Value_Values(SelectedID)
                    If Not Integer.TryParse(ParaString, ParaIndex) OrElse Line Is Nothing OrElse ParaIndex < 0 OrElse ParaIndex >= Line.Length Then
                        Final.Append("[")
                        Final.Append(ParaString)
                        Final.Append("]")
                    Else
                        Final.Append(Line(ParaIndex))
                    End If
                    ParaMode = False
                    Para = Nothing
                Else
                    Final.Append("]")
                End If
            Else
                If ParaMode Then
                    Para.Append(MaskArray(Index))
                Else
                    Final.Append(MaskArray(Index))
                End If
            End If
        Next
        If ParaMode Then
            Final.Append("[")
            Final.Append(Para.ToString)
        End If

        TextBox_Display.Text = Final.ToString
    End Sub

    ''' <summary>
    ''' Specified SelectedID with display text.
    ''' </summary>
    ''' <param name="ID"></param>
    ''' <param name="DisplayText"></param>
    ''' <remarks></remarks>
    Public Shadows Sub [Select](ByVal ID As Object, ByVal DisplayText As String)
        If ID <> Value_SelectedID Then
            Value_SelectedID = ID
            CloseSubWindow()
            RaiseEvent SelectedItemChanged(Me, New EventArgs)
        End If
        TextBox_Display.Text = DisplayText
    End Sub

    ''' <summary>
    ''' Specified SelectedID
    ''' </summary>
    ''' <param name="ID"></param>
    ''' <remarks></remarks>
    Public Shadows Sub [Select](ByVal ID As Object)
        If ID <> Value_SelectedID Then
            SelectedID = ID
            RaiseEvent SelectedItemChanged(Me, New EventArgs)
        End If
    End Sub

#End Region

#Region "Dropdown"

    Private SubWindow As Form_MultiColumnComboBoxDropDown

    Private Sub CloseSubWindow()
        If SubWindow IsNot Nothing Then
            RemoveHandler SubWindow.LostFocus, AddressOf SubWindowLostFocus
            SubWindow.Close()
            SubWindow = Nothing
        End If
    End Sub

    Public Class DropdownRequestEventArgs
        Inherits System.EventArgs
        Public Cancel As Boolean
    End Class

    Private Sub TextBox_Display_KeyDownOrMouseDown(ByVal sender As Object, ByVal e As EventArgs) Handles TextBox_Display.KeyDown, TextBox_Display.MouseClick
        Dim ea As New DropdownRequestEventArgs
        RaiseEvent DropdownRequested(Me, ea)
        If ea.Cancel Then Return
        If Value_Titles.Count = 0 Then Return
        If Value_Values.Count = 0 Then Return
        If SubWindow Is Nothing Then
            SubWindow = New Form_MultiColumnComboBoxDropDown
            SubWindow.ListView1.BeginUpdate()
            For Each item As TitleStructure In Value_Titles
                Dim h As New ColumnHeader
                h.Text = item.HeaderName
                h.Width = item.Width
                h.TextAlign = item.Align
                SubWindow.ListView1.Columns.Add(h)
            Next
            Dim Lines(Value_Values.Count - 1) As ListViewItem
            Dim LineIndex As Integer = 0
            For Each key As Object In Value_Values.Keys
                Lines(LineIndex) = New ListViewItem(Value_Values(key))
                If key = Value_SelectedID Then Lines(LineIndex).Selected = True
                Lines(LineIndex).Tag = key
                LineIndex += 1
            Next
            SubWindow.ListView1.Items.AddRange(Lines)
            SubWindow.ListView1.EndUpdate()
        End If

        Dim ps As Point = TextBox_Display.PointToScreen(New Point(TextBox_Display.Left, TextBox_Display.Top))
        WindowVisible.ShowWindow(SubWindow, ps.Y + TextBox_Display.Height, ps.X, ps.Y, Width, Value_DropDownHeight, Value_DropDownWidth)
        SubWindow.ListView1.Scrollable = True

        AddHandler SubWindow.Selected, AddressOf SubWindowSelectedChanged
        AddHandler SubWindow.ListView1.LostFocus, AddressOf SubWindowLostFocus
        SubWindow.ListView1.Focus()
    End Sub

    Private Sub SubWindowLostFocus(ByVal sender As Object, ByVal e As EventArgs)
        RemoveHandler SubWindow.Selected, AddressOf SubWindowSelectedChanged
        RemoveHandler SubWindow.ListView1.LostFocus, AddressOf SubWindowLostFocus
        SubWindow.ListView1.Scrollable = False
        SubWindow.ListView1.Visible = False
        WindowVisible.HideWindow(SubWindow)
        SubWindow.ListView1.Visible = True
    End Sub

    Private Sub SubWindowSelectedChanged()
        If SubWindow.ListView1.SelectedItems.Count < 1 Then Return
        Dim s As Object = SubWindow.ListView1.SelectedItems(0).Tag
        If s.Equals(SelectedID) = False Then
            Value_SelectedID = s
            ApplySelectText(s)
            RaiseEvent SelectedItemChanged(Me, New EventArgs)
            RaiseEvent SelectedItemChangedByUser(Me, New EventArgs)
        End If
        SubWindowLostFocus(Nothing, Nothing)
    End Sub

#End Region

#Region "Class: TitleItemCollection"

    Public Structure TitleStructure
        Public HeaderName As String
        Public Width As Integer
        Public Align As System.Windows.Forms.HorizontalAlignment
    End Structure

    Public Class TitleItemCollection
        Inherits List(Of TitleStructure)
        Friend Event ItemCollectionChanged()

        Public Shadows Sub Clear()
            MyBase.Clear()
            RaiseEvent ItemCollectionChanged()
        End Sub

        Public Shadows Sub Add(ByVal Item As TitleStructure)
            MyBase.Add(Item)
            RaiseEvent ItemCollectionChanged()
        End Sub

        Public Shadows Sub AddRange(ByVal Items As IEnumerable(Of TitleStructure))
            MyBase.AddRange(Items)
            RaiseEvent ItemCollectionChanged()
        End Sub

        Public Shadows Sub Insert(ByVal Index As Integer, ByVal Item As TitleStructure)
            MyBase.Insert(Index, Item)
            RaiseEvent ItemCollectionChanged()
        End Sub

        Public Shadows Sub InsertRange(ByVal Index As Integer, ByVal Items As IEnumerable(Of TitleStructure))
            MyBase.InsertRange(Index, Items)
            RaiseEvent ItemCollectionChanged()
        End Sub

        Public Shadows Sub RemoveAt(ByVal Index As Integer)
            MyBase.RemoveAt(Index)
            RaiseEvent ItemCollectionChanged()
        End Sub

        Public Shadows Sub Remove(ByVal Item As TitleStructure)
            MyBase.Remove(Item)
            RaiseEvent ItemCollectionChanged()
        End Sub

        Public Shadows Sub RemoveRange(ByVal Index As Integer, ByVal Count As Integer)
            MyBase.RemoveRange(Index, Count)
            RaiseEvent ItemCollectionChanged()
        End Sub

        Public Shadows Function RemoveAll(ByVal Match As Predicate(Of TitleStructure)) As Integer
            Dim rc = MyBase.RemoveAll(Match)
            RaiseEvent ItemCollectionChanged()
            Return rc
        End Function
    End Class

#End Region

#Region "Class: ValueItemCollection"
    Public Class ValueItemCollection
        Inherits Dictionary(Of Object, String())
        Friend Event ItemCollectionChanged()

        Public Shadows Sub Clear()
            MyBase.Clear()
            RaiseEvent ItemCollectionChanged()
        End Sub

        Public Shadows Sub Add(ByVal Key As Object, ByVal values As String())
            MyBase.Add(Key, values)
            RaiseEvent ItemCollectionChanged()
        End Sub

        Public Shadows Sub Remove(ByVal Key As Object)
            MyBase.Remove(Key)
            RaiseEvent ItemCollectionChanged()
        End Sub
    End Class

#End Region

    Private Sub Value_Titles_ItemCollectionChanged() Handles Value_Titles.ItemCollectionChanged, Value_Values.ItemCollectionChanged
        CloseSubWindow()
    End Sub

End Class